#---------------------------------------------------------------#
#           N-tap FIR filter                   			#
#---------------------------------------------------------------#
# * Function Description :                                      #
#   This routine performs a finite impulse respose filtering    #
#   with 16-bit coefficients and 16-bit inputs			#
#   The filter is discribed by the difference equation:         #
#          N-1                                                  #
#   Y(n) = SUM H(k)*X(n-k)                                      #
#          k=0                                                  #
#   where N is the number of coefficients.                      #
#                                                               #
# * PS :                                                        #
#   This is example program - 24 tap fir filter                 #
#                                                               #
# * Disign Assumption and Constraints :				#
#   If the number of tab not power of 2, then N is extended 	#
#   to be the power of 2 and 0's are put to the extended tap	#
#                                                               #
# * Input Values :						#
#   RAM0[0x80~ff] - input buffer         			#
# * Coefficients :                                              #
#   RAM1[40~5f] - fir filter coefficients H[k]			#
#                                                               #
# * Return Values :						#
#   RAM1[00~7f] - output value					#
#                                                               #
# * Altered Register :						#
#   Acc, ext7, x, y, stack1
# * Computation time :						#
#   maximum 48 cycles								#
#---------------------------------------------------------------#

org	0000
dw 	0004
dw 	0004
dw 	0004
dw 	0004
org	0004
     ## Initial part
     # set N selector
        ldi     (r3|01), 08 
     # If (0 < N <= 8)  ->  N selector=00
     # If (8 < N <= 16) ->  N selector=04
     # If (16 < N <= 32)  ->  N selector=08
     # If (32 < N <= 64)  ->  N selector=0c
     # If (64 < N <= 128)  ->  N selector=0f
     # set pointer
        ldi     r0, 40          # r0 : input buffer pointer
        ldi	r4, 40		# r4 : coeff. buffer pointer
        ldi     r5, 00          # r5 : output buffer pointer
     # load filter coefficients
        ldi	r4, 40		# initial address of coeff.
        ldi	a, @coeff_table
        ld	(r3|00), a
        ldi	a, 20		# counter : coeff. table size
     # N is changed to power of 2 for circular buffer
        ld	a, (r3|01)	# load N selector
        addi	a, @set_coeff_size
        ld	pc, a
     set_coeff_size:
        ldi	a, 08		# counter : coeff. table size (N=8)
        bra	always, @load_coeff
        ldi	a, 10		# counter : coeff. table size (N=16)
        bra	always, @load_coeff
        ldi	a, 20		# counter : coeff. table size (N=32)
        bra	always, @load_coeff
        ldi	a, 40		# counter : coeff. table size (N=64)
        bra	always, @load_coeff
        ldi	a, 80		# counter : coeff. table size (N=64)
     load_coeff:
        ld	stack, ((r3|00)) # load coeff. from P-Memory, (r3|00)++
        ld	(r4+), stack
        subi	1
        bra	z=0, @load_coeff
     # set loop size
        ld	a, (r3|01)
        addi    a, @set_loop_size
        ld	pc, a
     set_loop_size:
        ld	a, st
        ori	03		# loop size = 8  (N=8)
        bra	always, @save_st
        ld	a, st
        ori	04		# loop size = 16 (N=16)
        bra	always, @save_st
        ld	a, st
        ori	05		# loop size = 32 (N=32)
        bra	always, @save_st
        ld	a, st
        ori     06	        # loop size = 64 (N=64)
        bra	always, @save_st
        ld	a, st
        ori     06      	# loop size = 128 (N=128)
     save_st:
        ld  	st, a
        ldi	(r3|00), @input_table
        
     #-----------------------------------------------------
     ## filtering routines ( 16 sample output )
     fir_example():
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        call    always, @fir()
        ret
     
     fir():
     # input 1 sample
        ld      stack, ((r3|00)) # load input from p-memory, (r3|00)++
        ld 	(r0), stack      # put input to buffer
        call	always, @fir_sub
        ld      (r5+), a	 # save filtered output
        ret
        
     fir_sub:
        mld	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0)		# r0 is increased by 1
	add	a, p
	mod	always, shr
	ret

    coeff_table:
# N=24 -> N must be extended to 32 and 
# 0 must be inserted to N[24~31] for circular buffer
	dw      fff5		# N[0]
        dw      3               # N[1]
        dw      35              # N[2]
        dw      fff5            # N[3]
        dw      ff64            # N[4]
        dw      c               # N[5]
        dw      16a             # N[6]
        dw      20              # N[7]
        dw      fcdb            # N[8]
        dw      ff2e            # N[9]
        dw      f24             # N[10]
        dw      3b7             # N[11]
        dw      3b7             # N[12]
        dw      f24             # N[13]
        dw      ff2e            # N[14]
        dw      fcdb            # N[15]
        dw      20              # N[16]
        dw      16a             # N[17]
        dw      c               # N[18]
        dw      ff64            # N[19]
        dw      fff5            # N[20]
        dw      35              # N[21]
        dw      3               # N[22]
        dw      fff5            # N[23]
        dw      0               # N[24]
        dw      0               # N[25]
        dw      0               # N[26]
        dw      0               # N[27]
        dw      0               # N[28]
        dw      0               # N[29]
        dw      0               # N[30]
        dw      0               # N[31]
input_table:
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0001
        dw     0000
        dw     226a
        dw     c054
        dw     5364
        dw     a562
        dw     5442
        dw     beba
        dw     2482
        dw     fdba
        dw     dfb4
        dw     3e08
        dw     ad86
        dw     5a8e
        dw     aaee
        dw     42d6
        dw     d96a
        dw     048c
        dw     1e28
        dw     c3a6
        dw     5182
        dw     a58e
        dw     55d4
        dw     bba6
        dw     28a2
        dw     f92e
        dw     e400
        dw     3aa2
        dw     af84
        dw     5a46
        dw     a978
        dw     45d2
        dw     d558
