#---------------------------------------------------------------#
#   fir16_32.sc		                  			#
#---------------------------------------------------------------#
# * Function  :                                                 #
#   This routine performs a finite impulse respose filtering.   #
#   The filter is discribed by the difference equation:         #
#          N-1                                                  #
#   Y(n) = SUM H(k)*X(n-k)                                      #
#          k=0                                                  #
#   where N is the number of coefficients.                      #
#                                                               #
# * PS :                                                        #
#   This is example program - 24 tap fir filter                 #
#                                                               #
# * Disign Assumption and Constraints :				#
#   If N is not power of 2, N is extended to the size of power  #
#   of 2 and 0's are put to the extended tap. 			#
#                                                               #
# * Input Values :						#
#   RAM1[0x00~(0x07 or 0x0f or 0x1f or 0x3f or 0x7f] 		#
#   RAM1[0x80~(0x87 or 0x8f or 0x9f or 0xbf or 0xff] 		#
#    - filter coefficients					#
#   RAM0[0x80~(0x87 or 0x8f or 0x9f or 0xbf or 0xff] 		#
#    - input buffer         					#
# * Temporal Value :						#
#   RAM0[00] save progra addr.					#
#   RAM0[01] N selector                                         #
#   RAM0[10] sum of lower word mac				#
#                                                               #
# * Return Values :						#
#   RAM0[10~1f] - output value					#
#                                                               #
# * Computed time :						#
#   16 tap filtering - maximum 51 cycles			#
#---------------------------------------------------------------#

org	0000
dw 	0004
dw 	0004
dw 	0004
dw 	0004
org	0004
     ## Initial part
     # set pointer
        ldi     r0, 80          # r0 : input buffer pointer
        ldi	r4, 80	   # r4 : coeff. high word buffer pointer
        ldi	r5, 00	   # r5 : coeff. low word buffer pointer
        ldi     r2, 00          # r5 : output buffer pointer
     # set N selector
        ldi     (r3|01), 04
     # If (0 < N <= 8)  ->  N selector=00
     # If (8 < N <= 16) ->  N selector=04
     # If (16 < N <= 32)  ->  N selector=08
     # If (32 < N <= 64)  ->  N selector=0c
     # If (64 < N <= 128)  ->  N selector=0f
     
     # set output sample number
        ldi	r1, 20		# count for output samples
        
     # load filter coefficients
        ldi	r4, 80		# initial address of coeff.
        ldi	r5, 00
        ldi	a, @coeff_table
        ld	(r3|00), a
        ldi	a, 20		# counter : coeff. table size
     # N is changed to power of 2 for circular buffer
        ld	a, (r3|01)	# load N selector
        addi	a, @set_coeff_size
        ld	pc, a
        
     set_coeff_size:
        ldi	a, 08		# counter : coeff. table size (N=8)
        bra	always, @load_coeff
        ldi	a, 10		# counter : coeff. table size (N=16)
        bra	always, @load_coeff
        ldi	a, 20		# counter : coeff. table size (N=32)
        bra	always, @load_coeff
        ldi	a, 40		# counter : coeff. table size (N=64)
        bra	always, @load_coeff
        ldi	a, 80		# counter : coeff. table size (N=64)
     load_coeff:
        ld	stack, ((r3|00)) # load coeff. high word, (r3|00)++
        ld	(r4+), stack
        ld	stack, ((r3|00)) # load coeff. low word, (r3|00)++
        ld	(r5+), stack
        subi	1
        bra	z=0, @load_coeff
     
     # set loop size
        ld	a, (r3|01)
        addi    a, @set_loop_size
        ld	pc, a
     set_loop_size:
        ld	a, st
        ori	03		# loop size = 8  (N=8)
        bra	always, @save_st
        ld	a, st
        ori	04		# loop size = 16 (N=16)
        bra	always, @save_st
        ld	a, st
        ori	05		# loop size = 32 (N=32)
        bra	always, @save_st
        ld	a, st
        ori     06	        # loop size = 64 (N=64)
        bra	always, @save_st
        ld	a, st
        ori     06      	# loop size = 128 (N=128)
     save_st:
        ld  	st, a
        
        ldi	(r3|00), @input_table
        
     #-----------------------------------------------------
     ## filtering routine ( 16 sample output )
     fir16_32_example:
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        call	always, @fir16_32()
        ret
        
     fir16_32():
     # input 1 sample
        ld      stack, ((r3|00)) # load input from p-memory, (r3|00)++
        ld 	(r0), stack      # put input to buffer
        call	always, @fir16_32_sub
        ld      (r5+), a	 # save filtered output
        ld	(r5+), ext7
        ret
        
     fir16_32_sub:
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)
	mpya	(r5+), (r0-)		# r0 is increased by 1
	add	a, p
	mod	always, shr
	ld	(r3|10), a
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0-)
	mpya	(r4+), (r0)		# r0 is increased by 1
	ld	x, (r3|10)		# load lower word
	ldi	y, 0008			# shift right 12
	add	a, p
	ret

coeff_table:            
# N=24 -> N must be extended to 32 and 
# 0 must be inserted to N[24~31] for circular buffer
	dw      fff5		# N[0]
        dw	0555
        dw      3               # N[1]
        dw	0333
        dw      35              # N[2]
        dw	0555
        dw      fff5            # N[3]
        dw	0555
        dw      ff64            # N[4]
        dw	0444
        dw      c               # N[5]
        dw	0ccc
        dw      16a             # N[6]
        dw	0aaa
        dw      20              # N[7]
        dw	0000
        dw      fcdb            # N[8]
        dw	0bbb
        dw      ff2e            # N[9]
        dw	0eee
        dw      f24             # N[10]
        dw	0444
        dw      3b7             # N[11]
        dw	0777
        dw      3b7             # N[12]
        dw	0777
        dw      f24             # N[13]
        dw	0444
        dw      ff2e            # N[14]
        dw	0eee
        dw      fcdb            # N[15]
        dw	0bbb
        dw      20              # N[16]
        dw	0000
        dw      16a             # N[17]
        dw	0aaa
        dw      c               # N[18]
        dw	0ccc
        dw      ff64            # N[19]
        dw	0444
        dw      fff5            # N[20]
        dw	0555
        dw      35              # N[21]
        dw	0555
        dw      3               # N[22]
        dw	0333
        dw      fff5            # N[23]
        dw	0555
        dw      0               # N[24]
        dw	0
        dw      0               # N[25]
        dw	0
        dw      0               # N[26]
        dw	0
        dw      0               # N[27]
        dw	0
        dw      0               # N[28]
        dw	0
        dw      0               # N[29]
        dw	0
        dw      0               # N[30]
        dw	0
        dw      0               # N[31]
        dw	0
input_table:
        dw     19d4
        dw     c71e
        dw     4f6a
        dw     a5f6
        dw     572e
        dw     b8c0
        dw     2ca6
        dw     f4a6
        dw     e85e
        dw     3716
        dw     b1b6
        dw     59c2
        dw     a83a
        dw     48a4
        dw     d162
        dw     0d9c
        dw     156e
        dw     cabc
        dw     4d1e
        dw     a696
        dw     5850
        dw     b606
        dw     308e
        dw     f024
        dw     ecca
        dw     3368
        dw     b41a
        dw     5904
        dw     a734
        dw     4b44
        dw     cd8a
        dw     1218
        dw     10fa
        dw     ce7c
        dw     4aa2
        dw     a770
        dw     5938
        dw     b37c
        dw     3456
        dw     ebae
        dw     f144
        dw     2f98
        dw     b6b0
        dw     580c
        dw     a668
        dw     4db6
        dw     c9d2
        dw     1688
        dw     0c7c
