#---------------------------------------------------------------#
#   fir32_32.sc		                 			#
#---------------------------------------------------------------#
# * Function  :                                                 #
#   This routine performs a finite impulse respose filtering    #
#   with 32-bit filter coefficients and 32-bit inputs		#
#   The filter is discribed by the difference equation:         #
#          N-1                                                  #
#   Y(n) = SUM H(k)*X(n-k)                                      #
#          k=0                                                  #
#   where N is the number of coefficients.                      #
#                                                               #
# * PS :                                                        #
#   This is example program - 24 tap fir filter                 #
#                                                               #
# * Disign Assumption and Constraints :				#
#   If N is not power of 2, N is extended to the size of power  #
#   of 2 and 0's are put to the extended tap. 			#
#                                                               #
# * Input Values :						#
#   RAM1[0x00~(0x07 or 0x0f or 0x1f or 0x3f ]	 		#
#   RAM1[0x80~(0x87 or 0x8f or 0x9f or 0xbf ]			#
#    - filter coefficients					#
#   RAM0[0x80~(0x87 or 0x8f or 0x9f or 0xbf ]	 		#
#    - input buffer         					#
# * Temporal Value :						#
#   RAM0[00] save progra addr.					#
#   RAM0[01] N selector                                         #
#   RAM0[11] sum of lower word mac				#
#                                                               #
# * Return Values :						#
#   RAM0[10~3f] - output value					#
#                                                               #
# * Computed Cycles :						#
#   16 tap filtering - maximum 1239 cycles 			#
#---------------------------------------------------------------#

org	0000
dw 	0004
dw 	0004
dw 	0004
dw 	0004
org	0004
     ## Initial part
     # set N selector
        ldi     (r3|01), 04
     # If (0 < N <= 8)  ->  N selector=00
     # If (8 < N <= 16) ->  N selector=04
     # If (16 < N <= 32)  ->  N selector=08
     # If (32 < N <= 64)  ->  N selector=0c
     
     # set output sample number
        ldi	r1, 20		# count for output samples
        
     # load filter coefficients
        ldi	r4, 80		# initial address of coeff.
        ldi	a, @coeff_table
        ld	(r3|11), a
        ldi	a, 20		# counter : coeff. table size
     # N is changed to power of 2 for circular buffer
        ld	a, (r3|01)	# load N selector
        addi	a, @set_coeff_size
        ld	pc, a
        
     set_coeff_size:
        ldi	a, 08		# counter : coeff. table size (N=8)
        bra	always, @load_coeff
        ldi	a, 10		# counter : coeff. table size (N=16)
        bra	always, @load_coeff
        ldi	a, 20		# counter : coeff. table size (N=32)
        bra	always, @load_coeff
        ldi	a, 40		# counter : coeff. table size (N=64)
     load_coeff:
        ld	stack, ((r3|11)) # load coeff. high word, (r3|00)++
        ld	(r4+), stack
        ld	stack, ((r3|11)) # load coeff. low word, (r3|00)++
        ld	(r4+), stack
        subi	1
        bra	z=0, @load_coeff
     
     # set loop size
        ld	a, (r3|01)
        addi    a, @set_loop_size
        ld	pc, a
     set_loop_size:
        ld	a, st
        ori	04		# loop size = 16 (N=8)
        bra	always, @save_st
        ld	a, st
        ori	05		# loop size = 32 (N=16)
        bra	always, @save_st
        ld	a, st
        ori	06		# loop size = 64 (N=32)
        bra	always, @save_st
        ld	a, st
        ori     07	        # loop size = 128 (N=64)
        bra	always, @save_st
     save_st:
        ld  	st, a
     # set pointer
        ldi     r0, 80         # r0 : input buffer pointer
        ldi	r4, 80	   # r4 : coeff. high word buffer pointer
        ldi     r5, 40          # r5 : output buffer pointer
        
        ldi	(r3|11), @input_table
        
     #-----------------------------------------------------
     ## filtering routine ( 16 sample output )
     # 16 tap example
     fir32_32_example:
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        call	always, @fir32_32()
        ret
     
     fir32_32():
     # input 1 sample
        ld      stack, ((r3|11)) # load input from p-memory, (r3|00)++
        ld 	(r0+), stack      # put input to buffer
        ld      stack, ((r3|11)) # load input from p-memory, (r3|00)++
        ld 	(r0-), stack      # put input to buffer
        call	always, @fir_sub
        ld      (r5+), a	 # save filtered output
        ld	a, (r7|10)
        ld	(r5+), a
        ret
        
     fir_sub:
     	ldi	(r7|10), 00
     	ldi	(r7|11), 00
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        call	always, @mult_r_dp()
        call	always, @Add_r_dp2()
        ld 	-, (r0+)
        ld	-, (r0+)
	ret
	
#-------------------------------------------------------#
#   mult_r_dp.sc					#
#-------------------------------------------------------#
#   * Function:                                         #
#     32bit addition of two 32bit variables             #
#   * Design Assumption and Constraints:		#
#     - When overflow, return 2147483647                #
#  	   underflow, return -2147483648                #
#     - 30bit of the result is valid			#
#   * Input values                                      #
#          (r0) = X_L ( Lower of X )                	#
#          (r0+) = X_H ( Higher of X )              	#
#          (r4) = Y_L ( Lower of Y )                	#
#          (r4+) = Y_H ( Higher of Y )                  #
#   * Return values                                     #
#          RAM1[00] = Lower word of result              #
#          RAM1[01] = Higher word of result             #
#   * Temporal value                                    #
#          RAM0[00] = abs of X_L buffer ( Lower of X )  #
#          RAM0[01] = abs of X_H buffer ( Higher of X ) #
#          RAM1[00] = abs of Y_L buffer ( Lower of Y )  #
#          RAM1[01] = abs of Y_H buffer ( Higher of Y ) #
#	   RAM0[10] = Sign buffer                       #
#   * Altered Registers					#
#          Acc, r3, r7, stack, st                       #
#   * Computation cycles                                #
#          maximum 52 cycles                            #
#                                                       #
#-------------------------------------------------------#
    mult_r_dp():
	ld	a, st
	ori	a, 0200		# resop, st[9]=1 (mac shift on)
	ld	st, a
	
    # check sign of X*Y
     	mld	(r4), (r0)
     	ld	(r3|10), p
    # get absolute value of X, Y
        ld	a, (r4)         # check sign of Y
        and	a, -
        ld	a, (r4+)
        ld	ext7, (r4+)
        bra	n=0, @pos_Y
        mod	always, neg     # if Y<0, find -Y
    pos_Y:
        ld	(r7|01), a      # save -Y_H
        ld	(r7|00), ext7   # save -Y_L
        ld	a, (r7|00)
        mod     always, shr     # shift right to make unsigned value
        andi	a, 7fff         # disable sign bit
        ld	(r7|00), a      # save -Y_L >> 1
    check_sign:
        ld      a, (r0)          # check sign of X
        and     a, -
        ld      a, (r0+)
        ld      ext7, (r0-)
        bra     n=0, @pos_X
        mod     always, neg      # if X<0, find -X
    pos_X:
        ld      (r3|01), a       # save -X_H
        ld      (r3|00), ext7    # save -X_L
        ld      a, (r3|00)
        mod     always, shr      # shift right to make unsigned value
        andi    a, 7fff          # disable sign bit
        ld	(r3|00), a       # save -X_L >> 1
    cal_mult_r_dp():
	mld     (r7|00), (r3|00) # acc=0, p=(Right_low>>1)*(Left_low>>1)
	ld	stack, p         # higher 16bit p result is initial~
                                 # ~value of following mac operation
	mld 	(r7|00), (r3|01) # p=(Right_low>>1)*(Left_high)
	ld	ext7, stack
        mpya    (r7|01), (r3|00) # p=(Right_high)*(Left_low>>1)
    mid_sum:
        mpya    (r7|01), (r0-)   # dummy for (add a, p)
        ld	x, a
        ldi	y, 2             # shift right 15
        andi	00
    last_mult:
    	mod	f, setop
        mpya    (r7|01), (r3|01)  # p=(Right_high)*(Left_high)
        mpya    (r7|01), (r0-)    # dummy for (add a, p)
        ld	(r7|00), ext7     # save Result_Low
        ld	(r7|01), a	  # save Result_High
        ld      a, (r3|10)
    	and	a, -
        bra     n=0, @pos_out
        ld      ext7, (r7|00)     
    	ld	a, (r7|01)
        mod     always, neg       # if sign = negative -> neg. operation
        ld      (r7|00), ext7     # save Result_Low
        ld      (r7|01), a        # save Result_High
    pos_out:
    	ld	pc, stack

#-------------------------------------------------------#
#   add_r_dp2.sc					#
#     - input place : ram0 -> ram1			#
#-------------------------------------------------------#
#   * Function:                                         #
#     32bit addition of two 32bit variables             #
#   * Design Assumption and Constraints:		#
#     When overflow, return 2147483647                  #
#  	   underflow, return -2147483648                #
#   * Input values                                      #
#          RAM1[00] = X_L ( Lower of X )                #
#          RAM1[01] = X_H ( Higher of X )               #
#          RAM1[10] = Y_L ( Lower of Y )                #
#          RAM1[11] = Y_H ( Higher of Y )               #
#   * Return values                                     #
#          RAM1[00] = Lower word of result              #
#          RAM1[01] = Higher word of result             #
#   * Altered Registers					#
#          Acc, r3, r7, stack, st                       #
#   * Computation cycles                                #
#          maximum 20 cycles                            #
#                                                       #
#-------------------------------------------------------#
    Add_r_dp2():
      	mod	f, resop
      	ld	a, (r7|00)   		# Lower word of	X
        add     a, (r7|10)              # Lower word of Y
        ld	stack, a        	# save sum of lower word
      	ld	a, (r7|01)
      	bra	l=0, @Add_r_dp_higher
      	addi	1                       # add carry
    Add_r_dp_higher:
        mod	f, setop                # Higher word of X
        add	a, (r7|11)              # Higher word of Y
        bra	ov=1, @Add_r_dp_overf_sat
    Add_r_dp_save:
        ld	(r7|11), a           # save higher word of result
        ld	(r7|10), stack       # save lower word of result
        bra	always, @Add_r_dp_ret
    # Overflow Saturation !
    Add_r_dp_overf_sat:
        ld	(r7|11), a
        bra	n=0, @Add_r_dp_res_pos
        ld 	(r7|10), a		# saturation value : 0x7fffffff
        bra	always, @Add_r_dp_res_pos
    Add_r_dp_res_pos:
        ldi	(r7|10), 0000           # saturation value : 0x80000000
    Add_r_dp_ret:
        ld	pc, stack


coeff_table:            
# N=24 -> N must be extended to 32 and 
# 0 must be inserted to N[24~31] for circular buffer
	dw      fff5		# N[0]
        dw	0555
        dw      3               # N[1]
        dw	0333
        dw      35              # N[2]
        dw	0555
        dw      fff5            # N[3]
        dw	0555
        dw      ff64            # N[4]
        dw	0444
        dw      c               # N[5]
        dw	0ccc
        dw      16a             # N[6]
        dw	0aaa
        dw      20              # N[7]
        dw	0000
        dw      fcdb            # N[8]
        dw	0bbb
        dw      ff2e            # N[9]
        dw	0eee
        dw      f24             # N[10]
        dw	0444
        dw      3b7             # N[11]
        dw	0777
        dw      3b7             # N[12]
        dw	0777
        dw      f24             # N[13]
        dw	0444
        dw      ff2e            # N[14]
        dw	0eee
        dw      fcdb            # N[15]
        dw	0bbb
        dw      20              # N[16]
        dw	0000
        dw      16a             # N[17]
        dw	0aaa
        dw      c               # N[18]
        dw	0ccc
        dw      ff64            # N[19]
        dw	0444
        dw      fff5            # N[20]
        dw	0555
        dw      35              # N[21]
        dw	0555
        dw      3               # N[22]
        dw	0333
        dw      fff5            # N[23]
        dw	0555
        dw      0               # N[24]
        dw	0
        dw      0               # N[25]
        dw	0
        dw      0               # N[26]
        dw	0
        dw      0               # N[27]
        dw	0
        dw      0               # N[28]
        dw	0
        dw      0               # N[29]
        dw	0
        dw      0               # N[30]
        dw	0
        dw      0               # N[31]
        dw	0
input_table:
        dw     19d4
        dw      4444
        dw     c71e
        dw      eeee
        dw     4f6a
        dw 	aaaa
        dw     a5f6
        dw	6666
        dw     572e
        dw	eeee
        dw     b8c0
        dw	0000
        dw     2ca6
        dw	6666
        dw     f4a6
        dw	6666
        dw     e85e
        dw	eeee
        dw     3716
        dw	6666
        dw     b1b6
        dw	6666
        dw     59c2
        dw	2222
        dw     a83a
        dw	aaaa
        dw     48a4
        dw	4444
        dw     d162
        dw	2222
        dw     0d9c
        dw	cccc
        dw     156e
        dw	eeee
        dw     cabc
        dw	cccc
        dw     4d1e
        dw	eeee
        dw     a696
        dw	6666
        dw     5850
        dw	0000
        dw     b606
        dw	6666
        dw     308e
        dw	eeee
        dw     f024
        dw	4444
        dw     ecca
        dw	aaaa
        dw     3368
        dw	8888
        dw     b41a
        dw	aaaa
        dw     5904
        dw	4444
        dw     a734
        dw	4444
        dw     4b44
        dw	4444
        dw     cd8a
        dw	aaaa
        dw     1218
        dw     10fa
        dw     ce7c
        dw     4aa2
        dw     a770
        dw     5938
        dw     b37c
        dw     3456
        dw     ebae
        dw     f144
        dw     2f98
        dw     b6b0
        dw     580c
        dw     a668
        dw     4db6
        dw     c9d2
        dw     1688
        dw     0c7c
        dw     d25c
